from glta.remma.remma_cpu import remma_add, remma_dom, remma_epiAA_cpu, remma_epiAA_select_cpu, remma_epiAA_eff_cpu, remma_epiAA_pair_cpu, remma_random_pair, remma_epiAA_eff_cpu_c
from glta.remma.remma_gpu import *
